/*
 * NVRBMTrainer.cpp
 *
 *  Created on: 24-Jan-2009
 *      Author: spoon
 */

#include <cutil_inline.h>
#include <time.h>
#include <queue.h>
#include <WorkRequest.h>
#include <WorkResult.h>

#include "NVRBM.cuh"
#include "NVRBM_kernel.cuh"

using namespace std;

NVRBM::NVRBM(Matrix* visHid, Matrix* visHidInc, Matrix* hidBiases, Matrix* hidBiasInc,
             Matrix* visBiases, Matrix* visBiasInc, Matrix* visStdevs, Matrix* visStdevsInc,
             Matrix* miniNegDataPCD, Matrix* hidMeans, float epsW, float epsVisBias, float epsHidBias, float epsStdevs,
             float mom, float weightCost, float visStdev, float sparseCost, float sparseTarget,
             float sparseDamping, int numCD, int minibatchSize,
             bool learnStdevs, bool addGaussianNoise,
             int deviceID) : Thread(false) {

    this->numVis            = visHid->getNumRows();
    this->numHid            = visHid->getNumCols();

    this->hVisHid           = visHid;
    this->hVisHidInc        = visHidInc;
    this->hHidBiases        = hidBiases;
    this->hHidBiasInc       = hidBiasInc;
    this->hVisBiases        = visBiases;
    this->hVisBiasInc       = visBiasInc;
    this->hVisStdevs        = visStdevs;
    this->hVisStdevsInc     = visStdevsInc;
    this->hHidMeans         = hidMeans;

    this->hMiniNegDataPCD   = miniNegDataPCD;

    this->epsW              = epsW;
    this->epsVisBias        = epsVisBias;
    this->epsHidBias        = epsHidBias;
    this->epsStdevs         = epsStdevs;
    this->mom               = mom;
    this->weightCost        = weightCost;
    this->visStdev          = learnStdevs ? 1 : visStdev;
    this->binaryVis         = this->visStdev == 0;
    this->sparseCost        = sparseCost;
    this->sparseTarget      = sparseTarget;
    this->sparseDamping     = sparseDamping;
    this->numCD             = numCD;
    this->minibatchSize     = minibatchSize;

    this->learnStdevs       = learnStdevs;
    this->addGaussianNoise  = addGaussianNoise;

    this->deviceID          = deviceID;

    this->requestQueue      = new Queue<WorkRequest*>();
    this->resultQueue       = new Queue<WorkResult*>();

    this->data              = NULL; // The training data, initially null
}

/*
 * There's really no need for a destructor since this object has the lifetime
 * of the entire application.
 */
NVRBM::~NVRBM() {

}

void NVRBM::initCuda() {
    cudaSetDevice(deviceID < 0 ? cutGetMaxGflopsDeviceId() : deviceID);
    cublasInit();
    NVMatrix::initDeviceProps();
    NVMatrix::initRandom(time(0));
//    NVMatrix::initRandom(7);

    this->visHid            = new NVMatrix(*hVisHid, true);
    this->visHidInc         = new NVMatrix(*hVisHidInc, true);
    this->hidBiases         = new NVMatrix(*hHidBiases, true);
    this->hidBiasInc        = new NVMatrix(*hHidBiasInc, true);
    this->visBiases         = new NVMatrix(*hVisBiases, true);
    this->visBiasInc        = new NVMatrix(*hVisBiasInc, true);
    this->visStdevs         = learnStdevs ? new NVMatrix(*hVisStdevs, true) : NULL;
    this->visStdevsInc      = learnStdevs ? new NVMatrix(*hVisStdevsInc, true) : NULL;
    this->hidMeans          = sparseCost > 0 ? new NVMatrix(*hHidMeans, true) : NULL;

    this->visHid_T          = &this->visHid->getTranspose();

    this->posVisAct         = new NVMatrix(*visBiases);
    this->posHidAct         = new NVMatrix(*hidBiases);
    this->negVisAct         = new NVMatrix(*visBiases);
    this->negHidAct         = new NVMatrix(*hidBiases);

    this->hidProbs          = new NVMatrix(minibatchSize, numHid);
    this->miniNegData       = numCD == 0 ? new NVMatrix(*hMiniNegDataPCD, true) : new NVMatrix(minibatchSize, numVis);
    this->miniReconData     = numCD == 0 ? new NVMatrix(minibatchSize, numVis) : this->miniNegData;

    this->miniScaledData    = learnStdevs ? new NVMatrix(*miniNegData) : NULL;
    this->posStdevGrads     = learnStdevs ? new NVMatrix(*visBiases) : NULL;
    this->negStdevGrads     = learnStdevs ? new NVMatrix(*visBiases) : NULL;
    this->stdevGradsTmp1    = learnStdevs ? new NVMatrix(*miniNegData) : NULL;
    this->stdevGradsTmp2    = learnStdevs ? new NVMatrix(*miniNegData) : NULL;

    this->sparseGrads       = sparseCost > 0 ? new NVMatrix(*hidProbs) : NULL;
    this->sparseGradsSum    = sparseCost > 0 ? new NVMatrix(*hidBiases) : NULL;
}

void* NVRBM::run() {
    /*
     * This gets executed in a new CPU thread so it's ok to initialize cuda stuff here.
     */
    initCuda();

    while (true) {
        WorkRequest* req = requestQueue->dequeue();

        if (req->getRequestType() == WorkRequest::TRAIN) {
            loadNewData(req);
            train();
        } else if (req->getRequestType() == WorkRequest::TEST) {
            loadNewData(req);
            test();
        } else if (req->getRequestType() == WorkRequest::SYNC) {
            syncWithHost();
        }

        delete req;
    }
    return NULL;
}

void NVRBM::loadNewData(WorkRequest* req) {
    if (data == NULL) {
        data = new NVMatrix(*req->getHData(), true);
    } else {
        data->copyFromHost(*req->getHData());
    }
}

void NVRBM::syncWithHost() {
    visHid->copyToHost(*hVisHid);
    visHidInc->copyToHost(*hVisHidInc);
    hidBiases->copyToHost(*hHidBiases);
    hidBiasInc->copyToHost(*hHidBiasInc);
    visBiases->copyToHost(*hVisBiases);
    visBiasInc->copyToHost(*hVisBiasInc);

    if (numCD == 0)
        miniNegData->copyToHost(*hMiniNegDataPCD);

    if (learnStdevs) {
        visStdevs->copyToHost(*hVisStdevs);
        visStdevsInc->copyToHost(*hVisStdevsInc);
    }

    if (sparseCost > 0)
        hidMeans->copyToHost(*hHidMeans);

    resultQueue->enqueue(new SyncDoneResult());
}

void NVRBM::train() {
    int numMinibatches = (data->getNumRows() + minibatchSize - 1) / minibatchSize;
    double batchError = 0;
    for (int m = 0; m < numMinibatches; m++) {
        int miniStart = m * minibatchSize;
        int miniEnd = min(data->getNumRows(), (m + 1) * minibatchSize);
        NVMatrix* miniData = &data->sliceRows(miniStart, miniEnd); // view

        computeHidProbs(miniData);
        computeCDWeightGrads(miniData, true);

        if (numCD == 0)                     // PCD
            computeHidProbs(miniNegData);
        for (int c = 0; c < numCD; c++) {   // CD-x
            computeNegData(miniNegData); // This binarizes hidProbs
            computeHidProbs(miniNegData);
        }

        computeCDWeightGrads(miniNegData, false);
        batchError += computeError(miniData, miniNegData);

        updateWeights();

        if (numCD == 0) // New negative data for next PCD iteration
            computeNegData(miniNegData);

        // All these matrices are views
        delete miniData;
    }
    cudaThreadSynchronize();
    resultQueue->enqueue(new BatchDoneResult(batchError / data->getNumElements()));
}

/*
 * Should use a bigger minibatch size for test error computation, but that requires more memory.
 */
void NVRBM::test() {
    int numMinibatches = (data->getNumRows() + minibatchSize - 1) / minibatchSize;
    double batchError = 0;
    for (int m = 0; m < numMinibatches; m++) {
        int miniStart = m * minibatchSize;
        int miniEnd = min(data->getNumRows(), (m + 1) * minibatchSize);
        NVMatrix* miniData = &data->sliceRows(miniStart, miniEnd); // view

        computeHidProbs(miniData);

        for (int c = 0; c < max(1, numCD); c++) {
            computeNegData(miniReconData); // This binarizes hidProbs
            if (c < numCD - 1)
                computeHidProbs(miniReconData);
        }
        batchError += computeError(miniData, miniReconData);

        // All these matrices are views
        delete miniData;
    }
    cudaThreadSynchronize();
    resultQueue->enqueue(new BatchDoneResult(batchError / data->getNumElements()));
}

inline void NVRBM::computeHidProbs(NVMatrix* miniData) {
    NVMatrix* _miniData = miniData;
    if (learnStdevs) {
        miniData->eltWiseDivideByVector2(*visStdevs, *miniScaledData);
        _miniData = miniScaledData;
    }
    _miniData->rightMult(*visHid, !binaryVis ? 1 / visStdev : 1, *hidProbs);
    hidProbs->addVector(*hidBiases);
    hidProbs->apply(NVMatrix::LOGISTIC2);
}

inline void NVRBM::computeNegData(NVMatrix* out) {
    hidProbs->binarizeProbs();
    if (!binaryVis && learnStdevs) {
        hidProbs->rightMult(*visHid_T, *out);
        if (addGaussianNoise)
            out->addGaussianNoise();
        out->eltWiseMultByVector(*visStdevs);
    } else {
        if (!binaryVis && addGaussianNoise) {
            out->randomizeGaussian();
            out->addProduct(*hidProbs, *visHid_T,  visStdev, visStdev);
        } else {
            hidProbs->rightMult(*visHid_T, !binaryVis ? visStdev : 1, *out);
        }
    }
    out->addVector(*visBiases);
    if (binaryVis) {
        out->apply(NVMatrix::LOGISTIC2);
    }
}

inline void NVRBM::computeCDWeightGrads(NVMatrix* miniData, bool positivePhase) {
    hidProbs->sum(0, positivePhase ? *posHidAct : *negHidAct);
    miniData->sum(0, positivePhase ? *posVisAct : *negVisAct);

    NVMatrix* miniData_T = &miniData->getTranspose(); // view
    int numCases = miniData->getNumRows();
    if (positivePhase) {
//        miniData_T->rightMult(*hidProbs, *unitCorrs);
        visHidInc->addProduct(*miniData_T, *hidProbs, mom, epsW / numCases);
        if (sparseCost > 0) {
            hidMeans->add(*posHidAct, sparseDamping, (1 - sparseDamping) / numCases);
            kComputeSparseGrads<<<NUM_SPARSE_GRAD_BLOCKS,NUM_SPARSE_GRAD_THREADS_PER_BLOCK>>>(hidProbs->getDevData(),hidMeans->getDevData(),
                    sparseCost, sparseTarget, sparseGrads->getDevData(),hidProbs->getNumRows(),numHid);
            sparseGrads->sum(0, *sparseGradsSum);
            visHidInc->addProduct(*miniData_T, *sparseGrads, 1, -epsW / numCases);
        }
    } else {
        visHidInc->addProduct(*miniData_T, *hidProbs, 1, -epsW / numCases);
//        unitCorrs->addProduct(*miniData_T, *hidProbs,1,-1);
    }
    delete miniData_T;

    /*
     * This is disgusting but getting around it is even more so because
     * of the differing transposedness of miniData and all the other matrices.
     *
     * Seriously, need to transpose all my data someday.
     */
//    printf("learnst %d, epsstd %f\n", learnStdevs, epsStdevs); exit(0);
    if (learnStdevs && epsStdevs > 0 && !binaryVis) {
        /*
         * d(-F)/ds_i = ((v_i - b_i)^2 / s_i - v_i * sum_j {h_j * w_ij}) / s_i^2
         *                                           ^                     ^
         *                                           hidProbs * visHid_T   performed in updateWeights()
         */
        NVMatrix* stdevGrads = positivePhase ? posStdevGrads : negStdevGrads;
        hidProbs->rightMult(*visHid_T, *stdevGradsTmp1);
        stdevGradsTmp1->eltWiseMult(*miniData);
        miniData->addVector(*visBiases, -1, *stdevGradsTmp2);
        stdevGradsTmp2->apply(NVMatrix::SQUARE);
        stdevGradsTmp2->eltWiseDivideByVector2(*visStdevs);
        stdevGradsTmp2->subtract(*stdevGradsTmp1);
        stdevGradsTmp2->sum(0, *stdevGrads);
    }
}

/*
 * The second parameter is here because it depends on whether
 * we're training with PCD.
 */
inline float NVRBM::computeError(NVMatrix* posData, NVMatrix* negData) {
//    posData->print(3,3);
//    printf("=\n");
    negData->squaredDiff(*posData, *this->miniReconData);
    return this->miniReconData->sum();
}

inline void NVRBM::updateWeights() {
    int numCases = miniNegData->getNumRows();
    if (weightCost > 0)
        visHidInc->add(*visHid, -weightCost * epsW / numCases);
    visBiasInc->addSum(*posVisAct, *negVisAct, mom, epsVisBias / numCases, -epsVisBias / numCases);

    hidBiasInc->addSum(*posHidAct, *negHidAct, mom, epsHidBias / numCases, -epsHidBias / numCases);
//    visHidInc->print(3,3);
    if (sparseCost > 0) // propagate sparseness deriv to biases slower than to weights
        hidBiasInc->add(*sparseGradsSum, -epsHidBias / numCases);

    visHid->add(*visHidInc);
    visBiases->add(*visBiasInc);
    hidBiases->add(*hidBiasInc);

    if(learnStdevs && epsStdevs > 0) {
        posStdevGrads->add(*negStdevGrads, epsStdevs / numCases, -epsStdevs / numCases);
        visStdevsInc->add(*posStdevGrads, mom, 1);

        visStdevs->add(*visStdevsInc);
//        visStdevs->print(3,3);
    }
//    delete visStdevs_T;
}
